#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/input.h>
#include <linux/kd.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>

#define MAX_VALUE      0xff

#define EVDEV          "/dev/input/event"
#define PROGNAME       "key2xd"
#define VERSION        "0.1"

//#define DEBUG

#ifdef DEBUG
   #define dmsg( x, args... )   fprintf( stderr, x, ## args );
#else
   #define dmsg( x, y... )         /* */
#endif /* DEBUG */

#define IS_CMD( x, y ) strncasecmp( x, y, strlen( y ) ) == 0

typedef enum xcommand_enum {
   nothing    = 0,
   mousemove  = 1,
   mousermove = 2,
   mousedown  = 3,
   mouseup    = 4,
   mouseclick = 5,
   key        = 6,
   keydown    = 7,
   keyup      = 8,
   ssleep     = 9,
   uusleep    = 10
} xcmd;

typedef struct xcommand_struct {
   xcmd cmd;
   int val1;
   int val2;
} xcmd_struct;


typedef struct xcommand_seq {
   int entries;
   xcmd_struct *xcmdstruct;
} xcmd_seq;

static xcmd_seq keyTable[ MAX_VALUE ];

static Display *dpy = NULL;

static int mousexmove = 5;
static int mouseymove = 5;
static int mousexrmove = 20;
static int mouseyrmove = 20;
/* prototypes */
static int setup_key_table(char *mapFile);
static int setup_event_dev(void);
static void process_key(int efd);
static KeyCode thing_to_keycode( char *thing );
static void send_key( int keycode );
static void mouse_click( int button );
static void mouse_rel_move( int x, int y, int repeat );
static void mouse_move( int x, int y );
static void process_command( xcmd_struct *xcmdstruct, int repeat );
static int check_command( char * cmd, xcmd_struct *xcmdstruct );

static int check_command( char * cmd, xcmd_struct *xcmdstruct ) 
{
   /* Process a command */
   int rc = 1;
   int tmpx,tmpy;

   tmpx = tmpy = -99;
   if( IS_CMD( cmd, "mouseclick " ) ) {
      sscanf( cmd, "mouseclick %d", &tmpx );
      if( tmpx == 1 || tmpx == 2 ) {
         xcmdstruct->val1 = tmpx;
         xcmdstruct->cmd = mouseclick;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "mousedown " ) ) {
      sscanf( cmd, "mousedown %d", &tmpx );
      if( tmpx == 1 || tmpx == 2 ) {
         xcmdstruct->val1 = tmpx;
         xcmdstruct->cmd = mousedown;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "mouseup " ) ) {
      sscanf( cmd, "mouseup %d", &tmpx );
      if( tmpx == 1 || tmpx == 2 ) {
         xcmdstruct->val1 = tmpx;
         xcmdstruct->cmd = mouseup;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "mousemove " ) ) {
      sscanf( cmd, "mousemove %d %d", &tmpx, &tmpy );
      if( tmpx >= 0 && tmpy >= 0 ) {
         xcmdstruct->val1 = tmpx;
         xcmdstruct->val2 = tmpy;
         xcmdstruct->cmd = mousemove;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "mousermove " ) ) {
      sscanf( cmd, "mousermove %d %d", &tmpx, &tmpy );
      if( tmpx >= -1 && tmpy >= -1 && tmpx <=1 && tmpy <=1 ) {
         xcmdstruct->val1 = tmpx;
         xcmdstruct->val2 = tmpy;
         xcmdstruct->cmd = mousermove;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "key " ) ) {
      xcmdstruct->val1 = thing_to_keycode( cmd + 4 );
      if( xcmdstruct->val1 > 0 ) {
         xcmdstruct->cmd = key;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "keydown " ) ) {
      xcmdstruct->val1 = thing_to_keycode( cmd + 8 );
      if( xcmdstruct->val1 > 0 ) {
         xcmdstruct->cmd = keydown;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "keyup " ) ) {
      xcmdstruct->val1 = thing_to_keycode( cmd + 6 );
      if( xcmdstruct->val1 > 0 ) {
         xcmdstruct->cmd = keyup;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "sleep " ) ) {
      sscanf( cmd, "sleep %d", &tmpx );
      if( xcmdstruct->val1 >= 0 ) {
         xcmdstruct->cmd = ssleep;
         xcmdstruct->val1 = tmpx;
         rc = 0;
      }
   }
   else if( IS_CMD( cmd, "usleep " ) ) {
      sscanf( cmd, "usleep %d", &tmpx );
      if( xcmdstruct->val1 >= 0 ) {
         xcmdstruct->cmd = ssleep;
         xcmdstruct->val1 = tmpx;
         rc = 0;
      }
   }

   if(rc)
      fprintf( stderr, "Unknown command '%s'\n", cmd );
   else
      fprintf( stdout, "  Cmd: %d, Val: %d - %d\n", xcmdstruct->cmd, xcmdstruct->val1, xcmdstruct->val2 );

   return rc;
}


static int setup_key_table(char *mapFile)
{
   int i;
   FILE *fd;
   char line[1024];

   for( i = 0; i <= MAX_VALUE; i ++ ) {
      keyTable[i].entries = 0;
      keyTable[i].xcmdstruct = NULL;
   }

   fd = fopen(mapFile, "r");
   if( fd == NULL ) {
      fprintf(stderr, "Cannot open <%s>\n", mapFile);
      return 1;
   }
   memset(line, 0, sizeof(line));
   while( !feof(fd) ) {
      line[0] = '\0';
      fgets(line, sizeof(line)-1, fd);
      // delete trailing LF
      if( line[ strlen(line)-1 ] == '\n' )
         line[ strlen(line)-1 ] = '\0';

      if( strlen(line) > 9 && line[0] == '0' && line[1] == 'x' && line[4] == '=' ) {
         int keycode;
         if( line[3] >= 'a' )
            keycode = 10 + (line[3] - 'a');
         else
            keycode = line[3] - '0';

         if( line[2] >= 'a' )
            keycode += (10 + (line[2] - 'a'))*16;
         else
            keycode += (line[2] - '0')*16;


         if( keycode > 0 && keycode <= MAX_VALUE ) {
            int i;
            char *p = strchr(line+5, ';');

            fprintf(stdout, "Line <%s>\n", line);
            keyTable[keycode].entries = 1;
            // Count entries ( must be ; seperated )
            while( p != NULL && strlen(p) > 4 ) {
               *p = '\0';
               keyTable[keycode].entries ++;
               p = strchr(p + 1, ';');
            }
            keyTable[keycode].xcmdstruct = calloc( sizeof(xcmd_struct), keyTable[keycode].entries );

            p = line + 5;

            for( i = 0; i < keyTable[keycode].entries; i ++ ) {
               if( check_command( p, &( keyTable[keycode].xcmdstruct[i] ) ) ) {
                  keyTable[i].entries = 0;
                  free( keyTable[i].xcmdstruct );
               }
               p += strlen(p) + 1;
            }
            //fprintf(stdout, "Key %x : <%s>\n", keycode, keyTable[keycode]);
         }
         else
            fprintf(stderr, "Unknown Key <%s>\n", line);
      }
      else {
         if( line[0] != '#' && line[0] != '\0' && line[0] != '\n' ) {
            int val = 0;
            char *p = strchr( line, '=' );
            if( p )
               val = atoi( p + 1 );

            if( val > 0 ) {
               if( IS_CMD( line, "mousexmove=" ) )
                  mousexmove = val;
               else if( IS_CMD( line, "mouseymove=" ) )
                  mouseymove = val;
               else if( IS_CMD( line, "mousexrmove=" ) )
                  mousexrmove = val;
               else if( IS_CMD( line, "mouseyrmove=" ) )
                  mouseyrmove = val;
               else
                  val = 0;
            }
            if(! val)
               fprintf(stderr, "Unknown line <%s>\n", line);
            else
               fprintf(stdout, "<%s> set to %d\n", line, val);
         }
      }
   }
   return 0;
}

static int setup_event_dev()
{
   int rfd = -1;
   int i;
   char name[256];
   char info[128];

   for( i = 0; i < 16; i++ ) {
      sprintf( name, "%s%d", EVDEV, i );
      if( (rfd = open (name, O_RDONLY)) >= 0 ) {
         ioctl (rfd, EVIOCGNAME (sizeof (info)), info);
         if( strstr( info, "keyboard" ) ) {
            printf("Using: %s (%s)\n", name, info);
            break;
         }
         printf("Not using: %s (%s)\n", name, info);
         close( rfd );
         rfd = -1;
      }
      else
         fprintf(stderr,"Can't open input event device %s\n", name);
   }
   return rfd;
}


static void process_key(int efd)
{
   static int lastKey = 0;
   static struct timeval lastTime = {0,0};
   struct input_event event[64];
//    char message[1000];
   int len, idx, i;
   int repeat;


   repeat = 0;

   len = read(efd, event, sizeof(struct input_event) * 64);

   if( len < (int) sizeof(struct input_event) ) {
      fprintf(stderr, "Error processing input event %s\n", strerror(errno));
      return;
   }

   if( event[1].value < 1 || event[0].value > MAX_VALUE )
      return;

//    printf ("%d.%06d\n", lastTime.tv_sec, lastTime.tv_usec);
   idx = event[0].value;

   if( keyTable[idx].entries == 1 && lastKey == idx ) {
      int tdiff = event[0].time.tv_sec - lastTime.tv_sec;
      if( tdiff < 2 ) {
         if( (tdiff * 1000000) + event[0].time.tv_usec - lastTime.tv_usec < 500000 )
            repeat = 1;
      }
   }
   for( i = 0; i < keyTable[idx].entries > 0; i++ ) {
      process_command( &(keyTable[idx].xcmdstruct[i]), repeat );
   }
   lastKey = idx;
   lastTime = event[0].time;
   dmsg("%d %d.%06d %02x 0x%04x %04x\n", repeat, (int)event[0].time.tv_sec, (int)event[0].time.tv_usec, 
        event[0].type, event[0].value, event[0].code);
   return;
}

static KeyCode thing_to_keycode( char *thing ) {
   KeyCode kc;
   KeySym ks;

   ks = XStringToKeysym( thing );
   if( ks == NoSymbol ) {
      fprintf( stderr, "Unable to resolve keysym for '%s'\n", thing );
      return(KeyCode)0;
   }

   kc = XKeysymToKeycode( dpy, ks );
   dmsg( "String '%s' maps to keysym '%d'\n", thing, (int)ks );
   dmsg( "String '%s' maps to keycode '%d'\n", thing, (int)kc );

   return( kc );
}

static void send_key( int keycode ) {
   dmsg( "Sending key '%d'\n", keycode );

   XTestFakeKeyEvent( dpy, keycode, True, CurrentTime );
   XTestFakeKeyEvent( dpy, keycode, False, CurrentTime );
}

static void mouse_click( int button ) {
   dmsg( "Clicking mouse button %d\n", button );
   XTestFakeButtonEvent( dpy, button, True, CurrentTime );
   XTestFakeButtonEvent( dpy, button, False, CurrentTime );
}

static void mouse_move( int x, int y ) {
   dmsg( "Moving mouse to %d,%d\n", x, y );
   XTestFakeMotionEvent( dpy, -1, x, y, CurrentTime );
}

static void mouse_rel_move( int x, int y, int repeat ) {
   dmsg( "Moving mouse relatively by %d,%d\n", x, y );
   /* This does not match my docs... hrm... */
   if(repeat) {
      x *= mousexrmove;
      y *= mouseyrmove;
   }
   else {
      x *= mousexmove;
      y *= mouseymove;
   }
   XTestFakeRelativeMotionEvent( dpy, x, y, CurrentTime );
}

static void process_command( xcmd_struct *xcmdstruct, int repeat ) {
   /* Process a command */
   switch( xcmdstruct->cmd ) {
      case mouseclick:
         mouse_click( xcmdstruct->val1 );
         break;
      case mousermove:
         mouse_rel_move( xcmdstruct->val1, xcmdstruct->val2, repeat );
         break;
      case mousemove:
         mouse_move( xcmdstruct->val1, xcmdstruct->val2 );
         break;
      case mousedown:
         XTestFakeButtonEvent( dpy, xcmdstruct->val1, True, CurrentTime );
         break;
      case mouseup:
         XTestFakeButtonEvent( dpy, xcmdstruct->val1, False, CurrentTime );
         break;
      case key:
         send_key( xcmdstruct->val1 );
         break;
      case keydown:
         XTestFakeKeyEvent( dpy, xcmdstruct->val1, True, CurrentTime );
         break;
      case keyup:
         XTestFakeKeyEvent( dpy, xcmdstruct->val1, False, CurrentTime );
         break;
      case ssleep:
         sleep( xcmdstruct->val1 );
         break;
      case uusleep:
         usleep( xcmdstruct->val1 );
         break;
      default:
         fprintf( stderr, "Unknown command '%d'\n", xcmdstruct->cmd );
   }

   XFlush( dpy );
}

int main(int argc, char **argv)
{
   fd_set permset, fdset;
   int max_fd, event_fd;
   char *displayname;

   if( argc != 2 ) {
      fprintf(stderr,"Usage: %s <configFile>\n", argv[0] );
      return 1;
   }

   displayname = getenv("DISPLAY");

   dpy = XOpenDisplay( displayname ? displayname : ":0.0" );
   if( dpy == NULL ) {
      fprintf( stderr, "Unable to open display :0\n" );
      exit( 1 );
   }

   if( setup_key_table(argv[1]) ) {
      fprintf(stderr,"Usage: %s <configFile>\n", argv[0] );
      return 1;
   }

   printf( "\n%s Version %s\n", PROGNAME, VERSION );
   event_fd = setup_event_dev();
   if( event_fd < 0 ) {
      fprintf(stderr,"No input device found\n" );
      return 1;
   }

   signal(SIGPIPE, SIG_IGN);

   FD_ZERO(&permset);
   FD_SET(event_fd, &permset);
   max_fd = event_fd;
   max_fd++;

   while( 1 ) {
      fdset = permset;

      if( select(max_fd, &fdset, NULL, NULL, NULL) < 0 ) {
         if( errno == EINTR )
            continue;
         fprintf(stderr, "Error during select(): %s\n", strerror(errno));
         return -1;
      }

      if( FD_ISSET(event_fd, &fdset) )
         process_key(event_fd);
   }
   XCloseDisplay( dpy );
   return 0;
}


